#
# fsdbm.rb
#
# Copyright (c) 2003 Minero Aoki <aamine@loveruby.net>
#
# This program is free software.
# You can distribute/modify this program under the terms of
# the GNU Lesser General Public License version 2 or later.
#

require 'find'

class FSDBM

  # turn on tracing only when used with refe
  begin
    require 'refe/traceutils'
    include ReFe::TraceUtils
  rescue LoadError
    def trace( msg )
    end
  end

  def initialize( basedir )
    @basedir = basedir
    @base_re = %r<\A#{@basedir}/>
  end

  def []( *keys )
    path = keys_to_path(keys)
    if FileTest.file?(path)
      File.open(path) {|f|
          return f.read
      }
    elsif FileTest.directory?(path)
      return nil unless File.file?(path + '/-')
      File.open(path + '/-') {|f|
          return f.read
      }
    else
      nil
    end
  end

  def []=( *args )
    val = args.pop
    setup_dirs args[0..-2]
    begin
      File.open(keys_to_path(args), 'w') {|f|
          f.write val
      }
    rescue Errno::EISDIR
      File.open(keys_to_path(args) + '/-', 'w') {|f|
          f.write val
      }
    end
    val
  end

  def setup_dirs( dirs )
    return if dirs.empty?

    dir = @basedir.dup
    while d = dirs.shift
      dir << '/' << encode(d)
      begin
        Dir.mkdir(dir) unless FileTest.directory?(dir)
      rescue Errno::EEXIST
        File.rename dir, File.dirname(dir) + "/tmp#{$$}"
        Dir.mkdir dir
        File.rename File.dirname(dir) + "/tmp#{$$}", dir + '/-'
      end
    end
  end
  private :setup_dirs

  def keys
    list = []
    Find.find(@basedir) do |path|
      list.push path if FileTest.file?(path)
    end
    list.map {|p| path_to_name(p) }
  end

  def expand_to_names( pattern )
    expand_to_pathes(pattern).map {|path| path_to_name(path) }
  end

  def expand_to_keys( pattern )
    expand_to_pathes(pattern).map {|path| path_to_keys(path) }
  end

  METACHARS = {
    '*' => '*',
    '?' => '?',
    '[' => '[',
    ']' => ']',
    '/' => '/'
  }

  def expand_to_pathes( orig )
    pattern = orig.gsub(%r<\\.|\W>n) {|s|
        if s.size == 1
        then METACHARS[s] || encode(s)
        else encode(s[1,1])
        end
    }
    trace "fsdbm expan: dir = #{@basedir}"
    trace "fsdbm expan: pattern = '#{pattern}'"
    result = Dir.glob(@basedir + '/' + pattern).reject {|path|
                 File.directory?(path) and not File.file?(path + '/-')
             }
    trace "fsdbm expan: #{result.size} matches"
    result
  end

  def path_to_keys( path )
    decode(path.sub(@base_re,'').sub(%r</-\z>,'')).split(%r</>)
  end

  def path_to_name( path )
    decode(path.sub(@base_re,'').sub(%r</-\z>,'').gsub(%r</>,''))
  end

  def keys_to_path( keys )
    @basedir + '/' + keys.map {|n| encode(n) }.join('/')
  end

  private

  def encode( str )
    str.gsub(/\W/n) {|s| sprintf('%%%02x', s[0]) }
  end

  def decode( str )
    str.gsub(/%[\da-h]{2}/i) {|s| s[1,2].hex.chr }
  end

end
